'''This module contains the test cases for the LDA module.'''
import unittest
from chemfusekit.lda import LDASettings, LDA
from chemfusekit.lldf import LLDFSettings, LLDF

class TestLDA(unittest.TestCase):
    '''Test suite for the LDA module.'''

    def test_lda_settings(self):
        '''Test case against settings errors.'''
        # Check for negative component rejection
        with self.assertRaises(ValueError):
            LDASettings(components=-3, output=True)
        # Check if split tests with no output cause warnings:
        with self.assertRaises(Warning):
            LDASettings(output=False,test_split=True)

    def test_lda_constructor(self):
        '''Test case against constructor parameter issues.'''

        # Create an LLDF model and initialize it
        lldf_settings = LLDFSettings(
            qepas_path='tests/qepas.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rt.xlsx',
            rt_sheet='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(lldf_settings)
        lldf.lldf()

        lda_settings = LDASettings()

        # First, construct the object with null model:
        with self.assertRaises(TypeError):
            LDA(None, lda_settings)

        # Then, construct the object with null settings:
        with self.assertRaises(TypeError):
            LDA(lldf.fused_data, None)

        # Now, with both null:
        with self.assertRaises(TypeError):
            LDA(None, None)

        # Finally, with proper values:
        LDA(lldf.fused_data, lda_settings)
    
    def test_lda(self):
        '''Integration test case.''' 
        # Create an LLDF model and initialize it
        lldf_settings = LLDFSettings(
            qepas_path='tests/qepas.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rt.xlsx',
            rt_sheet='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(lldf_settings)
        lldf.lldf()

        # Create an LDA object and train it, with true output
        lda_settings = LDASettings(output=True)
        lda = LDA(lldf.fused_data, lda_settings)
        lda.lda()

        # Create an LDA object and train it, with false output
        lda_settings = LDASettings(output=False)
        lda = LDA(lldf.fused_data, lda_settings)
        lda.lda()

        # Create an LDA object and train it, with true output and split tests
        lda_settings = LDASettings(output=True, test_split=True)
        lda = LDA(lldf.fused_data, lda_settings)
        lda.lda()

    def test_lda_predict(self):
        '''Test case against prediction parameter issues.'''        

        # Create an LLDF model and initialize it
        lldf_settings = LLDFSettings(
            qepas_path='tests/qepas.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rt.xlsx',
            rt_sheet='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(lldf_settings)
        lldf.lldf()

        # Create an LDA object without training it
        lda_settings = LDASettings()
        lda = LDA(lldf.fused_data, lda_settings)

        # Pick a random sample for prediction
        x_data_sample = lldf.fused_data.x_train.iloc[119] # should be DMMP
        x_data_sample = x_data_sample.iloc[1:].to_frame().transpose()

        # Run prediction with untrained model (should throw exception)
        with self.assertRaises(RuntimeError):
            lda.predict(x_data_sample)

        # Train the LDA object
        lda.lda()

        # Run prediction with empty data (should throw exception)
        with self.assertRaises(TypeError):
            lda.predict(None)

        # Run prediction with trained model and non-null data
        lda.predict(x_data_sample)



'''This module contains the test cases for the PCA module.'''
import unittest
from chemfusekit.pca import PCASettings, PCA
from chemfusekit.lldf import LLDFSettings, LLDF

class TestPCA(unittest.TestCase):
    '''Test suite for the PCA module.'''

    def test_pca_settings(self):
        '''Test case against settings errors.'''
        # Test for rejection of a negative variance:
        with self.assertRaises(ValueError):
            PCASettings(target_variance=-0.3)

        # Test for rejection of an out-of-bounds confidence level:
        with self.assertRaises(ValueError):
            PCASettings(confidence_level=3.00)

        # Test for rejection of insufficient initial components:
        with self.assertRaises(ValueError):
            PCASettings(initial_components=2)

        # Now try with no mistakes:
        PCASettings(
            target_variance=0.98,
            confidence_level=0.9,
            initial_components=8,
            output=True
        )

    def test_pca_constructor(self):
        '''Test case against constructor parameter issues.'''

        # Create a LLDF model and initialize it
        lldf_settings = LLDFSettings(
            qepas_path='tests/qepas.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rt.xlsx',
            rt_sheet='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(lldf_settings)
        lldf.lldf()

        lda_settings = PCASettings()

        # First, construct the object with null model:
        with self.assertRaises(TypeError):
            PCA(None, lda_settings)

        # Then, construct the object with null settings:
        with self.assertRaises(TypeError):
            PCA(lldf.fused_data, None)

        # Now, with both null:
        with self.assertRaises(TypeError):
            PCA(None, None)

        # Finally, with proper values:
        PCA(lldf.fused_data, lda_settings)
    
    def test_pca(self):
        '''
        Integration test case to verify that the output does not change based on
        whether the output is set to true or false
        '''
        # Create a LLDF model and initialize it
        lldf_settings = LLDFSettings(
            qepas_path='tests/qepas.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rt.xlsx',
            rt_sheet='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(lldf_settings)
        lldf.lldf()

        # Set up and execute PCA
        pca_settings = PCASettings(output=True)
        pca = PCA(lldf.fused_data, pca_settings)
        pca.pca()

        # Save the results
        result_true_components = pca.components
        result_true_array_scores = pca.array_scores

        # Set up and execute PCA (again)
        pca_settings = PCASettings(output=False)
        pca = PCA(lldf.fused_data, pca_settings)
        pca.pca()

       # Save the results
        result_false_components = pca.components
        result_false_array_scores = pca.array_scores

        self.assertEqual(result_true_components, result_false_components)
        self.assertEqual(result_true_array_scores, result_false_array_scores)
 


'''Support Vector Machine module.'''
from typing import Optional

import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns

from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix, classification_report
from sklearn.svm import SVC

from chemfusekit.lldf import LLDFModel

class SVMSettings:
    '''Holds the settings for the SVM object.'''
    def __init__(self, kernel: str = 'linear', output: bool = False, test_split: bool = False):
        if kernel not in ['linear', 'poly', 'gaussian', 'sigmoid']:
            raise ValueError("Invalid type: must be linear, poly, gaussian or sigmoid")
        if test_split is True and output is False:
            raise Warning(
                "You selected test_split but it won't run because you disabled the output."
            )
        self.kernel = kernel
        self.output = output
        self.test_split = test_split

class SVM:
    '''Class for Support Vector Machine analysis of the data'''
    def __init__(self, fused_data: LLDFModel, settings: SVMSettings):
        self.fused_data = fused_data
        self.settings = settings
        self.model: Optional[SVC] = None

    def svm(self):
        '''Performs Support Vector Machine analysis'''

        x_data = self.fused_data.x_data
        x_train = self.fused_data.x_train
        y = self.fused_data.y

        x_train, x_test, y_train, y_test = train_test_split(
            x_data,
            y,
            train_size=0.7,
            shuffle=True,
            stratify=y
        )

        # Linear kernel
        if self.settings.kernel == 'linear':
            svm_model = SVC(kernel='linear', probability=True)
            # svmlinear.predict_proba(X_train)
        # Polynomial kernel
        elif self.settings.kernel == 'poly':
            svm_model = SVC(kernel='poly', degree=8)
        # Gaussian kernel - radial basis function
        elif self.settings.kernel == 'gaussian':
            svm_model = SVC(kernel='rbf')
        # Sigmoid classifier
        elif self.settings.kernel == 'sigmoid':
            svm_model = SVC(kernel='sigmoid')
        else:
            raise ValueError(f"SVM: this type of kernel does not exist ({self.settings.type=})")

        svm_model.fit(x_train, y_train)
        self.model = svm_model

        if self.settings.output:
            y_pred = svm_model.predict(x_test)

            # Assuming 'y_true' and 'y_pred' are your true and predicted labels
            cm = confusion_matrix(y_test, y_pred)

            # Get unique class labels from y_true
            class_labels = sorted(set(y_test))

            # Plot the confusion matrix using seaborn with custom colormap (Blues)
            sns.heatmap(
                cm,
                annot=True,
                fmt='d',
                cmap='Blues',
                xticklabels=class_labels,
                yticklabels=class_labels,
                cbar=False,
                vmin=0,
                vmax=cm.max()
            )

            plt.xlabel('Predicted')
            plt.ylabel('True')
            plt.title('Confusion Matrix based on evaluation set')
            plt.show()

            # Print the classification report
            print(classification_report(y_test, y_pred, digits=2))

    def predict(self, x_data: pd.DataFrame):
        '''Performs SVM prediction once the model is trained'''
        if self.model is None:
            raise RuntimeError("The model hasn't been trained yet!")
        if x_data is None:
            raise TypeError("X data for prediction cannot be empty.")

        return self.model.predict(x_data)



'''Principal Component Analysis Module'''
from typing import Optional

import numpy as np
import pandas as pd
import plotly.express as px
import matplotlib.pyplot as plt

from sklearn.decomposition import PCA as PC

import scipy.stats

from chemfusekit.lldf import LLDFModel

class PCASettings:
    '''Holds the settings for the PCA object.'''
    def __init__(self, target_variance: float = 0.95,
                 confidence_level: float = 0.05,
                 initial_components: int = 10, output: bool = False):
        if target_variance < 0:
            raise ValueError("Target variance should be positive or null.")
        if confidence_level < 0 or confidence_level > 1:
            raise ValueError("Confidence level should be between 0 and 1.")
        if initial_components < 3:
            raise ValueError("Initial components should be at least 3.")
        self.target_variance = target_variance
        self.confidence_level = confidence_level
        self.initial_components = initial_components
        self.output = output


class PCA:
    '''A class to store the data, methods and artifacts for Principal Component Analysis'''
    def __init__(self, fused_data: LLDFModel, settings: PCASettings):
        self.fused_data = fused_data
        self.components = 0
        self.pca_model: Optional[PC] = None
        self.settings = settings
        self.array_scores: Optional[np.ndarray] = None

    def pca(self):
        '''Performs Principal Component Analysis.'''
        # Read from the data fusion object
        x_data = self.fused_data.x_data

        # Run PCA producing the reduced variable Xreg and select the first 10 components
        pca = PC(self.settings.initial_components)
        pca.fit_transform(x_data)

        # Define the class vector (discrete/categorical variable)
        # y_dataframe = pd.DataFrame(self.fused_data.y, columns=['Substance'])
        # classes = y_dataframe.astype('category') (a cosa serve?)
        out_sum = np.cumsum(pca.explained_variance_ratio_)

        if self.settings.output:
            print("Proportion of Variance Explained : ", pca.explained_variance_ratio_)
            print("Cumulative Prop. Variance Explained: ", out_sum)

        # Autoselect the number of components
        for i,x in enumerate(out_sum):
            if x >= self.settings.target_variance:
                self.components = i
                break
        self.components = max(self.components, 3)

        if self.settings.output:
            print(f"Running PCA with {self.components} components")

            # PCA scree plot
            pc_values = np.arange(pca.n_components_) + 1
            plt.plot(pc_values, pca.explained_variance_ratio_, 'ro-', linewidth=2)
            plt.title('Scree Plot')
            plt.xlabel('Principal Component')
            plt.ylabel('Proportion of Variance Explained')
            plt.show()

            # Cumulative explained variance ratio
            plt.plot(pc_values, out_sum, 'ro-', linewidth=2)
            plt.title('Scree Plot')
            plt.xlabel('Principal Component')
            plt.ylabel('Cumulative Prop. Variance Explained')
            plt.show()

        # Run PCA producing the pca_model with a proper number of components
        pca = PC(n_components=self.components)
        self.pca_model = pca
        self.pca_model.fit_transform(x_data)


    def pca_stats(self):
        '''Produces PCA-related statistics.'''
        x_data = self.fused_data.x_data
        x_train = self.fused_data.x_train

        # Prepare the Scores dataframe (and concatenate the original 'Region' variable)
        pc_cols = [f"PC{i+1}" for i in range(self.components)]
        scores = pd.DataFrame(data=self.pca_model.fit_transform(x_data), columns=pc_cols)
        scores.index = x_data.index
        scores = pd.concat([scores, x_train.Substance], axis = 1)
        if self.settings.output:
            print(f"Scores:\n{scores}")

        # Prepare the loadings dataframe
        loadings = pd.DataFrame(
            self.pca_model.components_.T,
            columns=pc_cols,
            index=x_data.columns
        )
        loadings["Attributes"] = loadings.index
        if self.settings.output:
            print(f"Loadings:\n{loadings}")

        if self.settings.output:
            # View the scores plot using plotly library
            fig = px.scatter(
                scores,
                x="PC1",
                y="PC2",
                color="Substance",
                hover_data=['Substance'],
                hover_name=x_data.index
            )
            fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_layout(
                height=600,
                width=800,
                title_text='Scores Plot colored by Substance')
            fig.show()

            # Plot 3D scores
            fig = px.scatter_3d(
                scores,
                x='PC1',
                y='PC2',
                z='PC3',
                color='Substance',
                hover_data=['Substance'],
                hover_name=x_data.index
            )
            fig.show()

        # Get PCA scores
        t = scores.iloc[:,0:self.components]
        # Get PCA loadings
        p = loadings.iloc[:,0:self.components]
        # Calculate error array
        err = x_data - np.dot(t,p.T)
        # Calculate Q-residuals (sum over the rows of the error array)
        q= np.sum(err**2, axis=1)
        # Calculate Hotelling's T-squared (note that data are normalised by default)
        tsq = np.sum((t/np.std(t, axis=0))**2, axis=1)

        def mean_confidence_interval(data, confidence=self.settings.confidence_level):
            a = 1.0 * np.array(data)
            n = len(a)
            m, se = np.mean(a), scipy.stats.sem(a)
            h = se * scipy.stats.t.ppf((1 + confidence) / 2., n-1)
            return m, m-h, m+h

        tsq_conf = (mean_confidence_interval(
            tsq.values,
            confidence=self.settings.confidence_level)
        )[2]
        q_conf = (mean_confidence_interval(
            q.values,
            confidence=self.settings.confidence_level)
        )[2]

        # Create a dataframe using only T2 and Q-residuals
        hot_q_data = pd.DataFrame(
            {'T2': tsq, 'Qres': q, 'Substance': x_train.Substance},
            index = x_data.index
        )

        if self.settings.output:
            # Plot the Hotelling T2 vs Q-residuals plot
            fig = px.scatter(
                hot_q_data,
                x="T2",
                y="Qres",
                hover_data={'Sample': (hot_q_data.index)},
                color = "Substance"
            )
            fig.add_hline(y=abs(q_conf),line_dash="dot", line_color='Red')
            fig.add_vline(x=tsq_conf,line_dash="dot", line_color='Red')
            fig.update_traces(textposition='top center')
            fig.update_layout(
                height=600,
                width=800,
                title_text="Hotelling's T2 vs Q-residuals")
            fig.show()

        # Normalize Q-residuals and Hotelling's T-squared
        normalized_q = q / np.max(q)
        normalized_tsq = tsq / np.max(tsq)

        # Create a DataFrame with normalized values
        normalized_hot_q_data = {
            'T2': normalized_tsq,
            'Qres': normalized_q,
            'Substance': x_train.Substance
        }
        normalized_hot_q_data = pd.DataFrame(normalized_hot_q_data, index=x_data.index)

        if self.settings.output:
            # Plot the normalized Hotelling T2 vs Q-residuals plot
            fig_normalized = px.scatter(
                normalized_hot_q_data,
                x="T2",
                y="Qres",
                hover_data={'Sample': (normalized_hot_q_data.index)},
                color="Substance"
            )
            fig_normalized.add_hline(y=abs(q_conf / np.max(q)), line_dash="dot", line_color='Red')
            fig_normalized.add_vline(x=tsq_conf / np.max(tsq), line_dash="dot", line_color='Red')
            fig_normalized.update_traces(textposition='top center')
            fig_normalized.update_layout(
                height=600,
                width=800,
                title_text="Normalized Hotelling's T2 vs Q-residuals"
            )
            fig_normalized.show()

        # Assuming 'scores' is your DataFrame with the 'class' column
        # Drop the 'class' column before converting to NumPy array
        array_scores = scores.drop('Substance', axis=1).values

        if self.settings.output:
            print(f"Original DataFrame:\n{scores}")
            print(f"\nNumPy Array without 'Substance' column:\n{array_scores}")

        self.array_scores = array_scores



'''k-Nearest Neighbors Analysis module'''
from typing import Optional
from beartype.typing import Callable

from sklearn.neighbors import KNeighborsClassifier

import pandas as pd

from chemfusekit.lldf import LLDFModel
from chemfusekit.__utils import run_split_test, print_confusion_matrix

class KNNSettings:
    '''Holds the settings for the kNN object.'''
    def __init__(
            self,
            n_neighbors: int = 15,
            metric: str | Callable = 'euclidean',
            weights: str | Callable = 'uniform',
            algorithm: str = 'auto',
            output: bool = False,
            test_split: bool = False
        ):
        if n_neighbors < 1:
            raise ValueError("Invalid n_neighbors number: should be a positive integer.")
        if metric not in ['minkwoski', 'precomputed', 'euclidean'] and not callable(metric):
            raise ValueError(
                "Invalid metric: should be 'minkwoski', 'precomputed', 'euclidean' or a callable."
            )
        if weights not in ['uniform', 'distance'] and not callable(weights):
            raise ValueError("Invalid weight: should be 'uniform', 'distance' or a callable")
        if algorithm not in ['auto', 'ball_tree', 'kd_tree', 'brute']:
            raise  ValueError(
                "Invalid algorithm: should be 'auto', 'ball_tree', 'kd_tree' or 'brute'."
            )
        if test_split is True and output is False:
            raise Warning(
                "You selected test_split but it won't run because you disabled the output."
            )
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.weights = weights
        self.algorithm = algorithm
        self.output = output
        self.test_split = test_split

class KNN:
    '''Class to store the data, methods and artifacts for k-Nearest Neighbors Analysis'''
    def __init__(self, settings: KNNSettings, fused_data: LLDFModel):
        self.settings = settings
        self.fused_data = fused_data
        self.model: Optional[KNeighborsClassifier] = None

    def knn(self):
        '''Performs k-Nearest Neighbors Analysis'''
        # Prepare and train the kNN model
        knn = KNeighborsClassifier(
            n_neighbors=self.settings.n_neighbors,
            metric=self.settings.metric,
            weights=self.settings.weights,
            algorithm=self.settings.algorithm
        )
        knn.fit(self.fused_data.x_data, self.fused_data.y)

        # Save the trained model
        self.model = knn

        if self.settings.output:
            # View the prediction on the test data
            y_pred = knn.predict(self.fused_data.x_data)
            print(y_pred)

            print_confusion_matrix(
                self.fused_data.y,
                y_pred,
                "Confusion Matrix based on the whole data set"
            )

        if self.settings.test_split and self.settings.output:
            knn_split = KNeighborsClassifier(
                n_neighbors=self.settings.n_neighbors,
                metric=self.settings.metric,
                weights=self.settings.weights,
                algorithm=self.settings.algorithm
            )
            run_split_test(self.fused_data.x_data, self.fused_data.y, knn_split)


    def predict(self, x_data: pd.DataFrame):
        '''Performs kNN prediction once the model is trained.'''
        if x_data is None:
            raise TypeError("X data for kNN prediction must be non-empty.")
        if self.model is None:
            raise RuntimeError("The kNN model is not trained yet!")

        y_pred = self.model.predict(x_data)
        return y_pred



'''Logistic Regression Module'''
from typing import Optional

import numpy as np
import pandas as pd

from sklearn.linear_model import LogisticRegression

from chemfusekit.__utils import run_split_test, print_confusion_matrix

class LRSettings:
    '''Holds the settings for the LR object.'''
    def __init__(self, algorithm: str = 'liblinear', output: bool = False,
                 test_split: bool = False):
        if algorithm not in [
            'lbfgs',
            'liblinear',
            'newton-cg',
            'newton-cholesky',
            'sag',
            'saga'
        ]:
            raise ValueError(f"{algorithm}: this algorithm does not exist.")
        if test_split is True and output is False:
            raise Warning(
                "You selected test_split but it won't run because you disabled the output."
            )
        self.algorithm = algorithm
        self.output = output
        self.test_split = test_split

class LR:
    '''Class to store the data, methods and artifacts for Logistic Regression'''
    def __init__(self, settings: LRSettings, array_scores: np.ndarray, y: np.ndarray):
        self.settings = settings
        self.array_scores = array_scores
        self.y = y
        self.model: Optional[LogisticRegression] = None

    def lr(self):
        '''Performs Logistic Regression'''

        # Let's build our model on the training set
        model = LogisticRegression(
            solver=self.settings.algorithm,
            random_state=0,
            class_weight='balanced',
            max_iter=10000
        ).fit(self.array_scores, self.y)

        if self.settings.output:
            #we can see the classes the model used
            print(model.classes_)
            # See the intercept of the model
            print(model.intercept_)
            # See the coefficients of the model - that can be easily interpreted
            # (correlating or not with y)
            print(model.coef_)

        '''
        Evaluate the model: each sample has a probability of belonging to Positive
        or Negative outcome. Class 0 is Negative, class 1 is Positive.  If the value
        of the first column (probability of being Negative) is higher than 0.5, we
        have a Negative sample. Otherwise, it will be Positive
        '''

        probabilities = model.predict_proba(self.array_scores)
        predictions = model.predict(self.array_scores)

        # This tells us the accuracy of our model in calibration
        scores = model.score(self.array_scores, self.y)

        # Save the trained model
        self.model = model

        if self.settings.output:
            print(probabilities)
            print(predictions)
            print(scores)

            print_confusion_matrix(
                self.y,
                predictions,
                "Confusion Matrix based on whole data set",
            )

        if self.settings.test_split and self.settings.output:
            split_model = LogisticRegression(
                solver='lbfgs',
                random_state=0,
                class_weight='balanced',
                max_iter=10000
            )

            run_split_test(self.array_scores, self.y, split_model, extended=True)

    def predict(self, x_sample: pd.DataFrame):
        '''Performs LR prediction once the model is trained.'''
        if self.model is None:
            raise RuntimeError("The LR model is not trained yet!")

        prediction = self.model.predict(x_sample)
        probabilities = self.model.predict_proba(x_sample)

        if self.settings.output:
            print(prediction)
            print(probabilities)

        return prediction



'''Utilities model: functions that are shared between different classes'''
from sklearn.cross_decomposition import PLSRegression as PLSR
from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix, classification_report

import numpy as np

import plotly.express as px
import seaborn as sns
import matplotlib.pyplot as plt

def graph_output(scores, model, name: str):
    '''A reusable graphing function.'''
    print(scores)
    print(f"""
        explained variance ratio with {name}:
        {model.explained_variance_ratio_}
    """)

    # Display the explained variance ratio
    print("Explained Variance Ratio:", model.explained_variance_ratio_)

    #Scores plot
    fig = px.scatter(scores, x="LV1", y="LV2", color="Substance", hover_data=['Substance'])
    fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
    fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
    fig.update_layout(
        height=600,
        width=800,
        title_text='Scores Plot')
    fig.show()

    # Plot 3D scores
    fig = px.scatter_3d(scores, x='LV1', y='LV2', z='LV3',
                        color='Substance', hover_data=['Substance'],
                        hover_name=scores.index
    )
    fig.update_layout(title_text=f"3D colored by Substance for {name}")
    fig.show()

def run_split_test(x, y, model, extended=False):
    '''A function to run split tests on trained models.'''
    x_train, x_test, y_train, y_test = train_test_split(
        x,
        y,
        test_size=0.3,
        random_state=42,
        shuffle=True,
        stratify=y
    )

    model.fit(x_train, y_train)
    y_pred = model.predict(x_test)

    # TODO: add something to print _x_scores multimodally

    if extended:
        # We can see the classes the model used
        print(model.classes_)
        # See the intercept of the model
        print(model.intercept_)
        # See the coefficients of the model - that can be easily interpreted
        # (correlating or not with y)
        print(model.coef_)

        '''
        Evaluate the model: each sample has a probability of belonging to Positive
        or Negative outcome. Class 0 is Negative, class 1 is Positive.  If the value
        of the first column (probability of being Negative) is higher than 0.5, we
        have a Negative sample. Otherwise, it will be Positive
        '''
        probabilities = model.predict_proba(x_train)
        print(probabilities)

        # This tells us the accuracy of our model in calibration
        model.score(x_train, y_train)

        predictions = model.predict(x_train)

        print("Calibration predictions: ")
        print(predictions)
        print_confusion_matrix(y_train, predictions, "Confusion matrix based on training set")
 
    if isinstance(model, PLSR):
        y_pred = np.int8(np.abs(np.around(y_pred, decimals=0)))

    print_confusion_matrix(y_test, y_pred, "Confusion matrix based on evaluation set")

# TODO: make multimodal
def print_confusion_matrix(y1, y2, title):
    '''Function to simplify the plotting of confusion matrices'''
    cm = confusion_matrix(y1, y2)

    # Get unique class labels from y_true
    class_labels = sorted(set(y2))

    # Plot the confusion matrix using seaborn with custom colormap (Blues)
    sns.heatmap(cm,
        annot=True,
        fmt='d',
        cmap='Blues',
        xticklabels=class_labels,
        yticklabels=class_labels,
        cbar=False,
        vmin=0,
        vmax=cm.max()
    )

    plt.xlabel('Predicted')
    plt.ylabel('True')
    plt.title(title)
    plt.show()

    # Print the classification report
    print(classification_report(y1, y2, digits=2))



'''Linear Discriminant Analysis module'''
from typing import Optional

import numpy as np
import pandas as pd

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LD
from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix, classification_report

from chemfusekit.lldf import LLDFModel
from chemfusekit.__utils import graph_output, run_split_test

class LDASettings:
    '''Holds the settings for the LDA object.'''
    def __init__(self, components: int = 3, output: bool = False, test_split: bool = False):
        if components <= 2:
            raise ValueError("Invalid component number: must be a > 1 integer.")
        if test_split is True and output is False:
            raise Warning(
                "You selected test_split but it won't run because you disabled the output."
            )
        self.components = components
        self.output = output
        self.test_split = test_split

class LDA:
    '''Class to store the data, methods and artifacts for Linear Discriminant Analysis'''
    def __init__(self, lldf_model: LLDFModel, settings: LDASettings):
        self.settings = settings
        self.x_data = lldf_model.x_data
        self.x_train = lldf_model.x_train
        self.y = lldf_model.y
        self.model: Optional[LD] = None

    def lda(self):
        '''Performs Linear Discriminant Analysis'''

        lda = LD(n_components=self.settings.components) # N-1 where N are the classes
        scores_lda = lda.fit(self.x_data, self.y).transform(self.x_data)

        if self.settings.output:
            print(f"LDA scores:\n{scores_lda}")
            print(lda.priors_)
            print(lda.means_)
            print(lda.coef_[0:self.settings.components,:])
            pred=lda.predict(self.x_data)
            print(np.unique(pred, return_counts=True))

            print(confusion_matrix(pred, self.y))
            print(classification_report(self.y, pred, digits=2))

        lv_cols = [f'LV{i+1}' for i in range(self.settings.components)]
        scores = pd.DataFrame(data = scores_lda, columns = lv_cols) # latent variables
        scores.index = self.x_data.index
        y_dataframe = pd.DataFrame(self.y, columns=['Substance'])

        scores = pd.concat([scores, y_dataframe], axis = 1)

        # Store the traiend model
        self.model = lda

        # Show graphs if required by the user
        if self.settings.output:
            graph_output(scores, self.model, "Linear Discriminant Analysis")

            # Run split tests if required by the user
            if self.settings.test_split:
                run_split_test(
                    (scores.drop('Substance', axis=1).values),
                    self.y,
                    LD(n_components=self.settings.components)
                )

    def predict(self, x_data: pd.DataFrame):
        '''Performs LDA prediction once the model is trained.'''
        if x_data is None:
            raise TypeError("X data for LDA prediction must be non-empty.")
        if self.model is None:
            raise RuntimeError("The LDA model is not trained yet!")

        y_pred = self.model.predict(x_data)
        return y_pred



'''Performs low-level data fusion on input arrays, outputs the results'''
from typing import Optional

import numpy as np
import pandas as pd
from scipy.signal import savgol_filter

# COSE DA FARE: SCELTA PREPROCESSING, TESTARE EXPORT

class LLDFModel:
    '''Models the output data from the LLDF operation'''
    def __init__(self, x_data: pd.DataFrame, x_train: pd.DataFrame, y: np.ndarray):
        self.x_data = x_data
        self.x_train = x_train
        self.y = y

class LLDFSettings:
    '''Holds the settings for the LLDF object.'''
    def __init__(
        self,
        qepas_path: str,
        qepas_sheet: str,
        rt_path: str,
        rt_sheet: str,
        preprocessing: str = 'snv'
    ):
        if preprocessing not in ['snv', 'savgol', 'savgol+snv']:
            raise SyntaxError("This type of preprocessing does not exist.")
        self.qepas_path = qepas_path
        self.qepas_sheet = qepas_sheet
        self.rt_path = rt_path
        self.rt_sheet = rt_sheet
        self.preprocessing = preprocessing

class LLDF:
    '''Holds together all the data, methods and artifacts of the LLDF operation'''
    def __init__(self, settings: LLDFSettings):
        self.settings = settings
        self.fused_data: Optional[LLDFModel] = None

    def _snv(self, input_data: np.ndarray):
        '''Applies normalization to an input array'''
        # Define a new array and populate it with the corrected data
        output_data = np.zeros_like(input_data)
        for i in range(input_data.shape[0]):

            # Apply correction
            output_data[i,:] = (
                (input_data[i,:] - np.mean(input_data[i,:])) / np.std(input_data[i,:])
            )

        return output_data

    def lldf(self):
        '''Performs low-level data fusion'''
        try:
            spectra = pd.read_excel(
                self.settings.qepas_path,
                sheet_name=self.settings.qepas_sheet,
                index_col=0,
                header=0
            )

            ret_time = pd.read_excel(
                self.settings.rt_path,
                sheet_name=self.settings.rt_sheet,
                index_col=0,
                header=0
            )
        except Exception as exc:
            raise FileNotFoundError("Error opening the selected files.") from exc

        # select only numerical attributes
        x_spectra = spectra.iloc[:, 1:]
        x_time = ret_time.iloc[:, 1:]

        #Selection of classes from the spectra database
        y = spectra.loc[:, 'Substance'].values

        y_dataframe = pd.DataFrame(y, columns=['Substance'])

        # It is necessary to convert the column names as string to select them
        spectra.columns = spectra.columns.astype(str) # to make the colnames as text

        # Preprocessing
        if self.settings.preprocessing == 'snv':
            # Compute the SNV on spectra
            preprocessed_spectra = self._snv(x_spectra.values)
        elif self.settings.preprocessing == 'savgol':
            # Preprocessing with Savitzki-Golay
            # smoothing, defining the window, the order and the use of derivatives
            preprocessed_spectra = savgol_filter(x_spectra, 7, polyorder = 2, deriv=0)
        elif self.settings.preprocessing == 'savgol+snv':
            # We can also combine the preprocessing strategies together:
            # Savitzki-Golay - smoothing + SNV
            preprocessed_spectra = self._snv(savgol_filter(x_spectra, 7, polyorder = 2, deriv=0))
        else:
            raise SyntaxError(
                f"LLDF: this type of preprocessing does not exist ({self.settings.preprocessing=})"
            )


        # Create a new DataFrame with the processed numerical attributes
        processed_dataframe_spectra = pd.DataFrame(
            preprocessed_spectra,
            columns=spectra.columns[1:]
        )

        # Create a new DataFrame with the processed numerical attributes
        processed_dataframe_rt = pd.DataFrame(x_time.values, columns=ret_time.columns[1:])

        # X training set
        x_train = pd.concat(
            [y_dataframe, processed_dataframe_spectra, processed_dataframe_rt],
            axis = 1
        )

        # select only numerical attributes
        x_data = x_train.iloc[:, 1:]

        self.fused_data = LLDFModel(x_data, x_train, y)

    def export_data(self, export_path: str):
        '''Exports the data fusion artifacts to a file'''
        if self.fused_data is None:
            raise RuntimeError("Cannot export data before data fusion.")

        x_dataframe = pd.DataFrame(self.fused_data.x_data)
        x_train_dataframe = pd.DataFrame(self.fused_data.x_train)
        y_dataframe = pd.DataFrame(self.fused_data.y)

        try:
            pd.concat([x_train_dataframe, y_dataframe, x_dataframe], axis=1).to_excel(export_path)
        except Exception as exc:
            raise RuntimeError("Could not export data to the selected path.") from exc



'''This module contains the test cases for the SVM module.'''
import unittest
from chemfusekit.svm import SVMSettings, SVM
from chemfusekit.lldf import LLDFSettings, LLDF

class TestSVM(unittest.TestCase):
    '''Test suite for the LDA module.'''

    def test_svm_settings(self):
        '''Test case against settings errors.'''
        # Test against null type
        with self.assertRaises(TypeError):
            SVMSettings(None, False, False)
        # Test against null output selector
        with self.assertRaises(TypeError):
            SVMSettings('linear', None, False)
        # Test against null test_split selector
        with self.assertRaises(TypeError):
            SVMSettings('linear', None, False)
        # Test against non-existent kernels
        with self.assertRaises(TypeError):
            SVMSettings('linear', False, None)

        # Check if split tests with no output cause warnings:
        with self.assertRaises(Warning):
            SVMSettings(output=False,test_split=True)

        # Now call with proper values:
        SVMSettings(kernel='gaussian', output=True, test_split=False)


    def test_svm_constructor(self):
        '''Test case against constructor parameter issues.'''

        # Create an LLDF model and initialize it
        lldf_settings = LLDFSettings(
            qepas_path='tests/qepas.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rt.xlsx',
            rt_sheet='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(lldf_settings)
        lldf.lldf()

        svm_settings = SVMSettings()

        # First, construct the object with null model:
        with self.assertRaises(TypeError):
            SVM(None, svm_settings)

        # Then, construct the object with null settings:
        with self.assertRaises(TypeError):
            SVM(lldf.fused_data, None)

        # Now, with both null:
        with self.assertRaises(TypeError):
            SVM(None, None)

        # Finally, with proper values:
        SVM(lldf.fused_data, svm_settings)
    
    def test_svm(self):
        '''Integration test case.'''

        # Create an LLDF model and initialize it
        lldf_settings = LLDFSettings(
            qepas_path='tests/qepas.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rt.xlsx',
            rt_sheet='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(lldf_settings)
        lldf.lldf()

        # Create an SVM object and train it, with true output
        svm_settings = SVMSettings(output=True)
        svm = SVM(lldf.fused_data, svm_settings)
        svm.svm()


        # Create an SVM object and train it, with false output
        svm_settings = SVMSettings(output=True)
        svm = SVM(lldf.fused_data, svm_settings)
        svm.svm()

    def test_svm_predict(self):
        '''Test case against prediction parameter issues.'''        

        # Create an LLDF model and initialize it
        lldf_settings = LLDFSettings(
            qepas_path='tests/qepas.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rt.xlsx',
            rt_sheet='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(lldf_settings)
        lldf.lldf()

        # Create an SVM object without training it
        svm_settings = SVMSettings()
        svm = SVM(lldf.fused_data, svm_settings)

        # Pick a random sample for prediction
        x_data_sample = lldf.fused_data.x_train.iloc[119] # should be DMMP
        x_data_sample = x_data_sample.iloc[1:].to_frame().transpose()

        # Run prediction with untrained model (should throw exception)
        with self.assertRaises(RuntimeError):
            svm.predict(x_data_sample)

        # Train the SVM object
        svm.svm()

        # Run prediction with empty data (should throw exception)
        with self.assertRaises(TypeError):
            svm.predict(None)

        # Run prediction with trained model and non-null data
        svm.predict(x_data_sample)



'''This module contains the test cases for the LLDF module.'''
import unittest
from chemfusekit.lldf import LLDFSettings, LLDF

class TestLLDF(unittest.TestCase):
    '''Test suite for the LLDF module.'''

    def test_file_loading(self):
        '''Test case against file loading errors.'''
        # load a non-existent file on purpose
        settings = LLDFSettings(
            qepas_path='tests/notfound.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rtx.xlsx',
            rt_sheet='Sheet1',
            preprocessing='savgol'
        )
        lldf = LLDF(settings)
        self.assertRaises(FileNotFoundError, lldf.lldf)

    def test_preprocessing_techniques(self):
        '''Test case against wrong preprocessing user input.'''
        with self.assertRaises(SyntaxError):
            settings = LLDFSettings(
                qepas_path='tests/qepas.xlsx',
                qepas_sheet='Sheet1',
                rt_path='tests/rt.xlsx',
                rt_sheet='Sheet1',
                preprocessing='qpl' # test with non-existent preprocessing technique
            )
        # Now a correct value:
        settings = LLDFSettings(
                qepas_path='tests/qepas.xlsx',
                qepas_sheet='Sheet1',
                rt_path='tests/rt.xlsx',
                rt_sheet='Sheet1',
                preprocessing='snv' # test with non-existent preprocessing technique
            )

        # Make the value wrong and test LLDF
        lldf = LLDF(settings)
        lldf.settings.preprocessing='qpl' 
        self.assertRaises(SyntaxError, lldf.lldf)

        # now check if it works with the three real processing techniques:
        lldf.settings.preprocessing='snv'
        lldf.lldf()
        lldf.settings.preprocessing='savgol'
        lldf.lldf()
        lldf.settings.preprocessing='savgol+snv'
        lldf.lldf()
    
    def test_export(self):
        '''Test case against wrong export settings.'''
        settings = LLDFSettings(
            qepas_path='tests/qepas.xlsx',
            qepas_sheet='Sheet1',
            rt_path='tests/rt.xlsx',
            rt_sheet='Sheet1',
            preprocessing='snv' # test with non-existent preprocessing technique
        )
        lldf = LLDF(settings)
        
        # Try exporting data before data fusion
        with self.assertRaises(RuntimeError):
            lldf.export_data('path')

        # Perform data fusion
        lldf.lldf()

        # Try exporting data to an invalid path
        with self.assertRaises(RuntimeError):
            lldf.export_data('$£=0\//|') 

if __name__ == '__main__':
    unittest.main()



'''Partial Least Squares Discriminant Analysis module.'''
from typing import Optional

import pandas as pd
import numpy as np

import plotly.express as px

from sklearn.cross_decomposition import PLSRegression as PLSR

from chemfusekit.lldf import LLDFModel
from chemfusekit.__utils import print_confusion_matrix, run_split_test


class PLSDASettings:
    '''Holds the settings for the PLSDA object.'''
    def __init__(self, n_components: int = 3, output: bool = False, test_split: bool = False):
        if n_components < 1:
            raise ValueError("Invalid n_components number: should be a positive integer.")
        if test_split is True and output is False:
            raise Warning(
                "You selected test_split but it won't run because you disabled the output."
            )
        self.n_components = n_components
        self.output = output
        self.test_split = test_split

class PLSDA:
    '''
    Class to store the data, methods and artifacts for Partial Least Squares
    Discriminant Analysis
    '''
    def __init__(self, settings: PLSDASettings, fused_data: LLDFModel):
        self.settings = settings
        self.fused_data = fused_data
        self.model: Optional[PLSR] = None

    def plsda(self):
        '''Performs Partial Least Squares Discriminant Analysis'''
        x = self.fused_data.x_data
        y = self.fused_data.x_train.Substance.astype('category').cat.codes

        regr_pls = PLSR(n_components=self.settings.n_components)
        regr_pls.fit_transform(x,y)

        # Save the model
        self.model = regr_pls

        if self.settings.output:
            # Re-create the model
            regr_pls_2 = PLSR(n_components=self.settings.n_components)
            regr_pls_2.fit_transform(x,y)

            # Scores
            scores = regr_pls_2.x_scores_
            lv_cols = [f"LV{i+1}" for i in range(self.settings.n_components)]
            scores = pd.DataFrame(scores, columns = lv_cols)
            scores.index = self.fused_data.x_train.index
            y = self.fused_data.x_train.Substance
            scores = pd.concat([scores, y], axis = 1)
            print(scores)

            # Loadings
            loadings = pd.DataFrame(regr_pls.x_loadings_,columns = lv_cols)
            loadings["Attributes"] = self.fused_data.x_train.iloc[:,1:].columns
            print(loadings)

            # Scores plot
            fig = px.scatter(scores, x="LV1", y="LV2", color="Substance", hover_data=['Substance'])
            fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_layout(
                height=600,
                width=800,
                title_text='Scores Plot')
            fig.show()

            # Plot 3D scores
            fig = px.scatter_3d(scores, x='LV1', y='LV2', z='LV3',
                                color='Substance', hover_data=['Substance'],
                                hover_name=scores.index)
            fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_layout(
                title_text='Scores 3D Plot colored by Class using the raw data')
            fig.show()

            # Loadings plot
            fig = px.scatter(loadings, x="LV1", y="LV2",text="Attributes")
            fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_traces(textposition='top center')
            fig.update_layout(
                height=600,
                width=800,
                title_text='Loadings Plot')
            fig.show()


            # Predictions
            pred = regr_pls.predict(x)
            pred = np.int8(np.abs(np.around(pred, decimals=0)))
            print(f"Predicted data:\n\n{pred}")

            # Print confusion matrix
            y = self.fused_data.x_train.Substance.astype('category').cat.codes 
            print_confusion_matrix(y, pred, "Confusion Matrix based on training set")


        if self.settings.output and self.settings.test_split:
            x = self.fused_data.x_data
            y = self.fused_data.x_train.Substance.astype('category').cat.codes
            run_split_test(x, y, PLSR(self.settings.n_components))

    def predict(self, x_data: pd.DataFrame):
        '''Performs PLSDA prediction once the model is trained.'''
        if x_data is None:
            raise TypeError("X data for PLSDA prediction must be non-empty.")
        if self.model is None:
            raise RuntimeError("The PLSDA model is not trained yet!")

        y_pred = self.model.predict(x_data)
        return y_pred
